"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DistributedJob = void 0;
const assert_1 = __importDefault(require("assert"));
const nanoid_1 = require("nanoid");
const schummar_queue_1 = require("schummar-queue");
const helpers_1 = require("./helpers");
class DistributedJob {
    constructor(scheduler, collection, jobId, implementation, _a = {}) {
        var { maxParallel = DistributedJob.DEFAULT_MAX_PARALLEL, retryCount = scheduler.options.retryCount, retryDelay = scheduler.options.retryDelay, log = scheduler.options.log, lockDuration = scheduler.options.lockDuration, lockCheckInterval = scheduler.options.lockCheckInterval } = _a, otherOptions = __rest(_a, ["maxParallel", "retryCount", "retryDelay", "log", "lockDuration", "lockCheckInterval"]);
        this.scheduler = scheduler;
        this.collection = collection;
        this.jobId = jobId;
        this.implementation = implementation;
        this.hasShutDown = false;
        this.subscribedExecutionIds = new Array();
        this.options = Object.assign({ maxParallel, retryCount, retryDelay, log, lockDuration, lockCheckInterval }, otherOptions);
        this.q = new schummar_queue_1.Queue({ parallel: this.options.maxParallel });
        if (implementation) {
            this.schedule();
            this.checkLocks();
            this.next();
        }
    }
    execute(...[data, { delay = 0, executionId = (0, nanoid_1.nanoid)() } = {}]) {
        return __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            yield col.updateOne({
                jobId: this.jobId,
                executionId,
            }, {
                $setOnInsert: {
                    jobId: this.jobId,
                    executionId,
                    schedule: null,
                    nextRun: new Date(Date.now() + delay),
                    lock: null,
                    finished: null,
                    attempt: 0,
                    data: data !== null && data !== void 0 ? data : null,
                    progress: 0,
                    state: 'planned',
                },
            }, { upsert: true });
            return executionId;
        });
    }
    await(executionId) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                const listener = (job) => {
                    if (job.state === 'completed')
                        resolve(job.result);
                    else if (job.state === 'error')
                        reject(Error(job.error));
                    else
                        return;
                    this.subscribedExecutionIds = this.subscribedExecutionIds.filter((x) => x.listener !== listener);
                };
                this.subscribedExecutionIds.push({ executionId, listener });
                const col = yield this.collection;
                const existing = yield col.findOne({ jobId: this.jobId, executionId });
                if (existing)
                    listener(existing);
            }));
        });
    }
    executeAndAwait(...args) {
        return __awaiter(this, void 0, void 0, function* () {
            const id = yield this.execute(...args);
            return this.await(id);
        });
    }
    onProgress(executionId, callback) {
        let lastValue;
        const listener = (job) => {
            if (job.progress && job.progress !== lastValue)
                callback(job.progress);
            lastValue = job.progress;
            if (job.state === 'completed' || job.state === 'error')
                cancel();
        };
        const cancel = () => (this.subscribedExecutionIds = this.subscribedExecutionIds.filter((x) => x.listener !== listener));
        this.subscribedExecutionIds.push({ executionId, listener });
        (() => __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            const existing = yield col.findOne({ jobId: this.jobId, executionId });
            if (existing)
                listener(existing);
        }))();
        return cancel;
    }
    watch(executionId, callback) {
        const listener = (job) => {
            callback(job);
            if (job.state === 'completed' || job.state === 'error')
                cancel();
        };
        const cancel = () => (this.subscribedExecutionIds = this.subscribedExecutionIds.filter((x) => x.listener !== listener));
        this.subscribedExecutionIds.push({ executionId, listener });
        (() => __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            const existing = yield col.findOne({ jobId: this.jobId, executionId });
            if (existing)
                listener(existing);
        }))();
        return cancel;
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            this.hasShutDown = true;
            if (this.timeout) {
                clearTimeout(this.timeout.handle);
                delete this.timeout;
            }
            yield this.q.untilEmpty;
        });
    }
    schedule(lastJob) {
        return __awaiter(this, void 0, void 0, function* () {
            const { schedule } = this.options;
            const data = schedule && schedule.data;
            if (!schedule)
                return;
            const col = yield this.collection;
            yield col.updateOne({
                jobId: this.jobId,
                schedule: { $ne: null },
                state: 'planned',
                lock: lastJob ? null : undefined,
            }, {
                $setOnInsert: {
                    jobId: this.jobId,
                    executionId: (0, nanoid_1.nanoid)(),
                    schedule: schedule,
                    nextRun: (0, helpers_1.calcNextRun)(schedule, lastJob === null || lastJob === void 0 ? void 0 : lastJob.nextRun),
                    lock: null,
                    finishedOn: null,
                    attempt: 0,
                    data: data !== null && data !== void 0 ? data : null,
                    progress: 0,
                    state: 'planned',
                },
            }, { upsert: true });
        });
    }
    checkLocks() {
        return __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            while (!this.hasShutDown) {
                try {
                    const threshold = new Date(Date.now() - this.options.lockDuration);
                    const res = yield col.updateMany({ jobId: this.jobId, lock: { $lt: threshold } }, { $set: { lock: null } });
                    if (res.modifiedCount)
                        this.options.log('info', 'Unlocked jobs:', res.modifiedCount);
                }
                catch (e) {
                    this.options.log('warn', 'Failed to check locks:', e);
                }
                yield (0, helpers_1.sleep)(this.options.lockCheckInterval);
            }
        });
    }
    next() {
        if (this.hasShutDown || !this.implementation)
            return;
        this.q.clear(true);
        this.q.schedule(() => __awaiter(this, void 0, void 0, function* () {
            try {
                if (this.timeout) {
                    clearTimeout(this.timeout.handle);
                    delete this.timeout;
                }
                const col = yield this.collection;
                const now = new Date();
                const { value: job } = yield col.findOneAndUpdate({
                    jobId: this.jobId,
                    state: 'planned',
                    nextRun: { $lte: now },
                    lock: null,
                }, {
                    $set: { lock: now },
                });
                this.options.log('debug', 'find next', job === null || job === void 0 ? void 0 : job.executionId);
                if (!job) {
                    this.checkForNextRun();
                    return;
                }
                this.next();
                (0, assert_1.default)(this.implementation);
                (0, assert_1.default)(job.state === 'planned');
                try {
                    const q = new schummar_queue_1.Queue();
                    this.options.log('debug', 'execute next', job === null || job === void 0 ? void 0 : job.executionId);
                    const result = yield this.implementation(job.data, {
                        job,
                        setProgress: (progress) => q.schedule(() => this.setProgress(job, progress)),
                    });
                    yield q.untilEmpty;
                    this.options.log('debug', 'execute next done', job === null || job === void 0 ? void 0 : job.executionId);
                    if (job.schedule) {
                        yield this.schedule(job);
                    }
                    yield col.updateOne({ _id: job._id }, {
                        $set: {
                            lock: null,
                            finishedOn: new Date(),
                            state: 'completed',
                            result,
                            errror: null,
                        },
                    });
                }
                catch (e) {
                    const msg = e instanceof Error ? e.message : e instanceof Object ? JSON.stringify(e) : String(e);
                    const retry = job.attempt < this.options.retryCount;
                    yield col.updateOne({ _id: job._id }, {
                        $set: {
                            nextRun: retry ? new Date(Date.now() + this.options.retryDelay) : job.nextRun,
                            lock: null,
                            attempt: retry ? job.attempt + 1 : job.attempt,
                            progress: 0,
                            state: retry ? 'planned' : 'error',
                            error: msg,
                        },
                    });
                    throw e;
                }
            }
            catch (e) {
                this.options.log('error', 'next failed', e);
            }
        }));
    }
    setProgress(job, progress) {
        return __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            yield col.updateOne({ _id: job._id }, { $set: { lock: new Date(), progress } });
        });
    }
    checkForNextRun() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.hasShutDown || !this.implementation)
                return;
            const col = yield this.collection;
            const [next] = yield col
                .find({
                jobId: this.jobId,
                lock: null,
                state: 'planned',
            })
                .sort({ nextRun: 1 })
                .limit(1)
                .toArray();
            if (next)
                this.planNextRun(next);
        });
    }
    receiveUpdate(job) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const { executionId, listener } of this.subscribedExecutionIds) {
                if (executionId === job.executionId)
                    listener(job);
            }
            if (job.state === 'planned') {
                return this.planNextRun(job);
            }
        });
    }
    changeStreamReconnected() {
        var e_1, _a;
        return __awaiter(this, void 0, void 0, function* () {
            this.checkForNextRun();
            const executionIds = new Set(this.subscribedExecutionIds.map((x) => x.executionId));
            const col = yield this.collection;
            const cursor = col.find({ jobId: this.jobId, executionId: { $in: [...executionIds] } });
            try {
                for (var cursor_1 = __asyncValues(cursor), cursor_1_1; cursor_1_1 = yield cursor_1.next(), !cursor_1_1.done;) {
                    const job = cursor_1_1.value;
                    yield this.receiveUpdate(job);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (cursor_1_1 && !cursor_1_1.done && (_a = cursor_1.return)) yield _a.call(cursor_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
        });
    }
    planNextRun(job) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.hasShutDown || !this.implementation)
                return;
            const now = Date.now();
            const date = new Date(Math.min(job.nextRun.getTime(), now + 60 * 60 * 1000));
            if (!this.timeout || date.getTime() < this.timeout.date.getTime()) {
                this.options.log('debug', 'plan next run', date.toISOString());
                if (this.timeout)
                    clearTimeout(this.timeout.handle);
                this.timeout = {
                    handle: setTimeout(() => this.next(), date.getTime() - now),
                    date,
                };
            }
        });
    }
    getPlanned() {
        return __awaiter(this, void 0, void 0, function* () {
            const col = yield this.collection;
            return yield col
                .find({
                jobId: this.jobId,
                state: 'planned',
            })
                .toArray();
        });
    }
}
exports.DistributedJob = DistributedJob;
DistributedJob.DEFAULT_MAX_PARALLEL = 1;
